/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
 /* Author: Zeehasham Rasheed
	Computer Science Department
	George Mason University 
	Dated: 10/15/2012
*/

package gmu.cs.dm.pig.eval;

import java.io.IOException;
import java.util.*;

import gmu.cs.dm.sequence.SequenceString;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.pig.EvalFunc;
import org.apache.pig.FuncSpec;
import org.apache.pig.data.*;
import org.apache.pig.impl.logicalLayer.schema.Schema;
import org.apache.pig.impl.logicalLayer.FrontendException;

/**
 * This function divides a search query string into wrods and extracts
 * n-grams with up to _ngramSizeLimit length.
 * Example 1: if query = "a real nice query" and _ngramSizeLimit = 2,
 * the query is split into: a, real, nice, query, a real, real nice, nice query
 * Example 2: if record = (u1, h1, pig hadoop) and _ngramSizeLimit = 2,
 * the record is split into: (u1, h1, pig), (u1, h1, hadoop), (u1, h1, pig hadoop)
 */
public class CalculateMinwiseHash extends EvalFunc<DataBag> {
   private static final Log LOG = LogFactory.getLog(CalculateMinwiseHash.class );

   public DataBag exec(Tuple input) throws IOException
   {
      //Tuple output = TupleFactory.getInstance().newTuple(1);
		DataBag output = DefaultBagFactory.getInstance().newDefaultBag();
      if ((input == null) || (input.size() == 0))
      {
         return(null);
      }

      try{
	  
		DataBag Kmer  = (DataBag)input.get(0);
		DataBag SeqID = (DataBag)input.get(1);
		int NumHash = (Integer)input.get(2);
		long mydiv = (Integer)input.get(3);
		//int KmerSize = (Integer)input.get(2);
		//System.out.println(SeqID);
		//System.out.println(Kmer);
		//System.out.println(NumHash);
		long[] hashA = new long[NumHash];
		long[] hashB = new long[NumHash];
		long[] MinHashValue = new long[NumHash];
		int i;
		
		for (i=0; i<NumHash; i++){
			hashA[i] = i+1;
			hashB[i] = i;
			MinHashValue[i] = 99999999;
		}
		long hash_calc;
		int n;
		int j;
		for (n=0; n<NumHash; n++){
			for (Tuple t : Kmer) {
				//mydiv = t.size();
				for (j=0;j<t.size();j++){
				//hash_calc = ((hashA[n] * (Long)t.get(j) + hashB[n]) % mydiv) % (mydiv+1);
				hash_calc = ((hashA[n] * (Long)t.get(j) + hashB[n]) % mydiv);
				MinHashValue[n] = ((hash_calc < MinHashValue[n])? hash_calc : MinHashValue[n]);
				}
			}
		}
/*		long size=0;
        for (Tuple t : Kmer) {
			//for (int i=0;i<t.size();i++){
			//System.out.println(t.get(i));
			//System.out.println("Another\n");
			//}
			size = size + t.size();
		}
		System.out.println(size);

/*		Iterator it = Kmer.iterator();
		try{
			while (it.hasNext())
			{
				//Tuple t = (Tuple)it.next();
				//sum += (Long)t.get(0);
				System.out.println((Tuple)it.next());
				System.out.println("Another\n");
			}
		}catch (Exception e) {
			System.err.println("KmerGenerator: failed to process input; error - " + e.getMessage());}

		
		/*
        int SeqLength = Seq.length();
		 
		long[] SeqKmerRep = new long[SeqLength-KmerSize];

		long kmer_prod;
		long kmer_sum;
		int k;
		char mychar;
		long charint;
		Tuple t = DefaultTupleFactory.getInstance().newTuple(SeqLength-KmerSize);
		DataBag oo = DefaultBagFactory.getInstance().newDefaultBag();
         for (int i=0; i <SeqLength-KmerSize; i++)
         {
			String SubSeq = Seq.substring(i,i+KmerSize);
			
			for (kmer_prod=1,kmer_sum=0,k=0; k<KmerSize; k++){
			  mychar = SubSeq.charAt(k);
			  charint = mychar;
              kmer_sum += kmer_prod * (charint);
              kmer_prod = kmer_prod * 4;
            }
			SeqKmerRep[i] = kmer_sum;
			//Tuple t = DefaultTupleFactory.getInstance().newTuple(1);
			//t.set(i, kmer_sum);
		}
		Arrays.sort(SeqKmerRep);
		
*/		DataBag oo = DefaultBagFactory.getInstance().newDefaultBag();
		Tuple t = DefaultTupleFactory.getInstance().newTuple(MinHashValue.length);
		for (i=0; i <MinHashValue.length; i++){
			t.set(i, MinHashValue[i]);
		}
		oo.add(t);
		Tuple t2 = DefaultTupleFactory.getInstance().newTuple(2);
		t2.set(0,oo);
		
		//for(int i=0; i <SeqLength-KmerSize; i++){
		//System.out.println(SeqKmerRep[i]);
		//}
		
        //output.set(0, SeqKmerRep);
		
		//t.set(SeqLength-KmerSize, SeqID);
		//DataBag ooo = DefaultBagFactory.getInstance().newDefaultBag();
		//Tuple t3 = DefaultTupleFactory.getInstance().newTuple(1);
		//t3.set(0, SeqID);
		//ooo.add(t3);
		
		t2.set(1,SeqID);
		output.add(t2);
		
		return(output);
      }
      catch (Exception e) {
         System.err.println("CalculateMinwiseHash: failed to process input; error - " + e.getMessage());
         return(null);
      }
      
   }
   @Override
   public Schema outputSchema(Schema input)
   {
      try {
         Schema.FieldSchema tokenFs = new Schema.FieldSchema("seqkmer",
                                                             DataType.LONG);
         //Schema tupleSchema = new Schema(tokenFs);
		
		Schema tupleSchema = new Schema();
        //tupleSchema.add(input.getField(0));
        //tupleSchema.add(input.getField(1));
        tupleSchema.add(tokenFs);
        tupleSchema.add(input.getField(1));
		
         Schema.FieldSchema tupleFs;
         tupleFs = new Schema.FieldSchema("tuple_of_seq_kmer", tupleSchema,
                                          DataType.TUPLE);

         Schema bagSchema = new Schema(tupleFs);
         bagSchema.setTwoLevelAccessRequired(true);
         Schema.FieldSchema bagFs = new Schema.FieldSchema(
            "seqskmers", bagSchema, DataType.BAG);

         return(new Schema(bagFs));
      }
      catch (FrontendException e) {
         // throwing RTE because
         //above schema creation is not expected to throw an exception
         // and also because superclass does not throw exception
         throw new RuntimeException("Unable to compute TOKENIZE schema.");
      }
   }


}
