/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
  /* Author: Zeehasham Rasheed
	Computer Science Department
	George Mason University 
	Dated: 10/15/2012
*/


package gmu.cs.dm.pig.eval;

import java.io.IOException;
import java.util.*;


import gmu.cs.dm.sequence.SequenceString;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.pig.EvalFunc;
import org.apache.pig.FuncSpec;
import org.apache.pig.data.*;
import org.apache.pig.impl.logicalLayer.schema.Schema;
import org.apache.pig.impl.logicalLayer.FrontendException;

/**
 * This function divides a search query string into wrods and extracts
 * n-grams with up to _ngramSizeLimit length.
 * Example 1: if query = "a real nice query" and _ngramSizeLimit = 2,
 * the query is split into: a, real, nice, query, a real, real nice, nice query
 * Example 2: if record = (u1, h1, pig hadoop) and _ngramSizeLimit = 2,
 * the record is split into: (u1, h1, pig), (u1, h1, hadoop), (u1, h1, pig hadoop)
 */
public class CalculatePairwiseSimilarity extends EvalFunc<DataBag> {
   private static final Log LOG = LogFactory.getLog(CalculatePairwiseSimilarity.class );

   public DataBag exec(Tuple input) throws IOException
   {
		
      //Tuple output = TupleFactory.getInstance().newTuple(1);
		DataBag output = DefaultBagFactory.getInstance().newDefaultBag();
      if ((input == null) || (input.size() == 0))
      {
		
        return(null);
      }
		
	  int i,j;
	  long [][] MinHashValue;
	  long [] OneMinHash;
	  int count=0;
      try{
		//System.out.println("In Hierarchical Clustering");
		DataBag OneKmer = (DataBag)input.get(0);
		DataBag Kmer = (DataBag)input.get(1);
		int NumHash = (Integer)input.get(2);
		
		int NumSeq = (int)Kmer.size();
		
		MinHashValue = new long [NumSeq][NumHash];
		OneMinHash = new long [NumHash];
		
		//System.out.println(OneKmer);
		//System.out.println(OneKmer.size());
		//System.out.println(Kmer);
		//System.out.println(Kmer.size());
		
		//System.out.println(NumHash);
		//System.out.println(Threshold);
		
		for (Tuple t : Kmer) {
				for (j=0;j<t.size()-1;j++){
					//System.out.println(t.get(j));
					MinHashValue[count][j] = (Long)t.get(j);
				}
				count=count+1;
		}
/*		
		for (i=0;i<NumSeq;i++){
			for (j=0;j<NumHash;j++){
				System.out.print(MinHashValue[i][j] +" ");
			}
			System.out.println("One Record Ends\n");
		}	
*/		


		for (Tuple t : OneKmer) {
				for (j=0;j<t.size();j++){
					//System.out.println(t.get(j));
					OneMinHash[j] = (Long)t.get(j);
				}
		}		
		
		
		double [] sim = new double[NumSeq];
		int k;
		int intersections;
		int unions;
		int cno;
		double simval;
		int numcluster=0;
		
		for (i=0; i < NumSeq; i++) {
			for (intersections =0, unions=0, k=0; k < NumHash; k ++){
					intersections += ((OneMinHash[k] == MinHashValue[i][k]) ? 1:0);            
					unions ++;
			}
			simval = 1.0*intersections/unions;
			sim[i] = simval;
		}
		
		
		
/*		
		cno = 0;
		for (i=0; i < NumSeq; i++){
			if(sim[i]!=0){}
			else{
			cno += 1;
			sim[i] = cno;
		
			for (j=0; j < NumSeq; j++) {
			  if (sim[j]!=0){}
			  else{
				for (intersections =0, unions=0, k=0; k < NumHash; k ++){
					intersections += ((MinHashValue[i][k] == MinHashValue [j][k]) ? 1:0);            
					unions ++;
					}
				simval = 1.0*intersections/unions;
				if(simval >= Threshold){
					  sim[j] = cno;
					  }
				}
			}
		  }
		}
		
		Tuple t = DefaultTupleFactory.getInstance().newTuple(sim.length+1);
		t.set(0, cno);
		
		for (i=0; i <sim.length; i++){
			t.set(i+1, sim[i]);
		}
*/		
		
/*		
		DataBag oo = DefaultBagFactory.getInstance().newDefaultBag();
		Tuple t = DefaultTupleFactory.getInstance().newTuple(MinHashValue.length);
		for (i=0; i <MinHashValue.length; i++){
			t.set(i, MinHashValue[i]);
		}
		oo.add(t);
		Tuple t2 = DefaultTupleFactory.getInstance().newTuple(2);
		t2.set(0,oo);
		
		
		t2.set(1,SeqID);
		output.add(t2);
		
		return(output);
*/      
		Tuple t = DefaultTupleFactory.getInstance().newTuple(sim.length);
		for (i=0; i <sim.length; i++){
			t.set(i, sim[i]);
		}
		output.add(t);
		return(output);
		
		}
      catch (Exception e) {
         System.err.println("CalculatePairwiseSimilarity: failed to process input; error - " + e.getMessage());
         return(null);
      }
      
   }
   /*
   @Override
   public Schema outputSchema(Schema input)
   {
      try {
         Schema.FieldSchema tokenFs = new Schema.FieldSchema("labels",
                                                             DataType.INTEGER);
         //Schema tupleSchema = new Schema(tokenFs);
		
		Schema tupleSchema = new Schema();
        //tupleSchema.add(input.getField(0));
        //tupleSchema.add(input.getField(1));
        tupleSchema.add(tokenFs);
        tupleSchema.add(input.getField(1));
		
         Schema.FieldSchema tupleFs;
         tupleFs = new Schema.FieldSchema("tuple_of_seq_kmer", tupleSchema,
                                          DataType.TUPLE);

         Schema bagSchema = new Schema(tupleFs);
         bagSchema.setTwoLevelAccessRequired(true);
         Schema.FieldSchema bagFs = new Schema.FieldSchema(
            "seqskmers", bagSchema, DataType.BAG);

         return(new Schema(bagFs));
      }
      catch (FrontendException e) {
         // throwing RTE because
         //above schema creation is not expected to throw an exception
         // and also because superclass does not throw exception
         throw new RuntimeException("Unable to compute TOKENIZE schema.");
      }
   }
*/
/*
public Schema outputSchema(Schema input) {
   try{
      Schema.FieldSchema counter = new Schema.FieldSchema("clusterlabel", DataType.INTEGER);
      Schema tupleSchema = new Schema(input.getField(1).schema.getField(1).schema.getFields());
      tupleSchema.add(counter);

      Schema.FieldSchema tupleFs;
      tupleFs = new Schema.FieldSchema("tuple", tupleSchema, DataType.TUPLE);

      Schema bagSchema = new Schema(tupleFs);
      return new Schema(new Schema.FieldSchema("bag",
                                                bagSchema, DataType.BAG));
     }catch (Exception e){
        return null;
     }
  } 
*/
}