/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
  /* Author: Zeehasham Rasheed
	Computer Science Department
	George Mason University 
	Dated: 10/15/2012
*/

package gmu.cs.dm.pig.eval;

import java.io.IOException;
import java.util.*;

import edu.ucla.sspace.common.Similarity.SimType;
import edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.ClusterLinkage;
import edu.ucla.sspace.matrix.*;
import edu.ucla.sspace.vector.*;
import edu.ucla.sspace.clustering.Merge;
import edu.ucla.sspace.util.*;

import gmu.cs.dm.sequence.SequenceString;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.pig.EvalFunc;
import org.apache.pig.FuncSpec;
import org.apache.pig.data.*;
import org.apache.pig.impl.logicalLayer.schema.Schema;
import org.apache.pig.impl.logicalLayer.FrontendException;

/**
 * This function divides a search query string into wrods and extracts
 * n-grams with up to _ngramSizeLimit length.
 * Example 1: if query = "a real nice query" and _ngramSizeLimit = 2,
 * the query is split into: a, real, nice, query, a real, real nice, nice query
 * Example 2: if record = (u1, h1, pig hadoop) and _ngramSizeLimit = 2,
 * the record is split into: (u1, h1, pig), (u1, h1, hadoop), (u1, h1, pig hadoop)
 */
public class CutoffHierarchicalClustering extends EvalFunc<DataBag> {
   private static final Log LOG = LogFactory.getLog(CutoffHierarchicalClustering.class );

   public DataBag exec(Tuple input) throws IOException
   {
      //Tuple output = TupleFactory.getInstance().newTuple(1);
		DataBag output = DefaultBagFactory.getInstance().newDefaultBag();
      if ((input == null) || (input.size() == 0))
      {
         return(null);
      }
		
	  int i,j;
	  double [][] MinHashValue;
	  int [] ClusterResults;
	  int count=0;
/*	  double[][] VALUES = {
        {2, 4, 5},
        {0, 0, 1},
        {5, 0, 1},
      };
*/	
      try{
		
		DataBag MinHash = (DataBag)input.get(0);
		int Option = (Integer)input.get(1);
		int NumHash = (Integer)input.get(2);
		double Cutoff = (Double)input.get(3);
			
		//System.out.println(SimMatrix);
		
		int NumSeq = (int)MinHash.size();
		
		MinHashValue = new double [NumSeq][NumHash];
		//Matrix matrix = new OnDiskMatrix(NumSeq,NumHash);
		
		
		System.out.println("Size of Bag");
		System.out.println(NumSeq);
		System.out.println("Cutoff");
		System.out.println(Cutoff);
		//System.out.println(m.rows());
		//System.out.println(m.columns());
		//System.out.println(m.get(2,1));
		
		for (Tuple t : MinHash) {
				for (j=0;j<t.size()-1;j++){
					//System.out.println(t.get(j));
					MinHashValue[count][j] = (Long)t.get(j);
					//matrix.set(count,j,(Long)t.get(j));
				}
				count=count+1;
		}	
		System.out.println("Matrix Made");
		Matrix matrix = new ArrayMatrix(MinHashValue);
		System.out.println("Matrix Conversion Done");
		
		if (Option==1)
		ClusterResults = edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.clusterRows(matrix,Cutoff,ClusterLinkage.COMPLETE_LINKAGE,SimType.JACCARD_INDEX);
		else if (Option==2)
		ClusterResults = edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.clusterRows(matrix,Cutoff,ClusterLinkage.MEAN_LINKAGE,SimType.JACCARD_INDEX);
		else if (Option==3)
		ClusterResults = edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.clusterRows(matrix,Cutoff,ClusterLinkage.MEDIAN_LINKAGE,SimType.JACCARD_INDEX);
		else if (Option==4)
		ClusterResults = edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.clusterRows(matrix,Cutoff,ClusterLinkage.SINGLE_LINKAGE,SimType.JACCARD_INDEX);
		else 
		ClusterResults = edu.ucla.sspace.clustering.HierarchicalAgglomerativeClustering.clusterRows(matrix,Cutoff,ClusterLinkage.COMPLETE_LINKAGE,SimType.JACCARD_INDEX);
		//System.out.println(ClusterResults);
		
		Tuple t = DefaultTupleFactory.getInstance().newTuple(ClusterResults.length);
		
		for (i=0; i <ClusterResults.length; i++){
			t.set(i, ClusterResults[i]);
			//System.out.println(ClusterResults[i]);
		}
		
/*		String s = "" + ClusterResults;
		Tuple t = DefaultTupleFactory.getInstance().newTuple(1);
		//t.set(0, clusterToElements);
		t.set(0, ClusterResults);
		output.add(t);
		return(output);
*/		
		output.add(t);
		return(output);
		//return(null);
		}
      catch (Exception e) {
         System.err.println("CutoffHierarchicalClustering: failed to process input; error - " + e.getMessage());
         return(null);
      }
      
   }
/*
   @Override
   public Schema outputSchema(Schema input)
   {
      try {
         Schema.FieldSchema tokenFs = new Schema.FieldSchema("dendrogram",
                                                             DataType.CHARARRAY);
         //Schema tupleSchema = new Schema(tokenFs);
		
		Schema tupleSchema = new Schema();
        //tupleSchema.add(input.getField(0));
        //tupleSchema.add(input.getField(1));
        tupleSchema.add(tokenFs);
		
         Schema.FieldSchema tupleFs;
         tupleFs = new Schema.FieldSchema("tuple_of_dendrogram", tupleSchema,
                                          DataType.TUPLE);

         Schema bagSchema = new Schema(tupleFs);
         bagSchema.setTwoLevelAccessRequired(true);
         Schema.FieldSchema bagFs = new Schema.FieldSchema(
            "seqskmers", bagSchema, DataType.BAG);

         return(new Schema(bagFs));
      }
      catch (FrontendException e) {
         // throwing RTE because
         //above schema creation is not expected to throw an exception
         // and also because superclass does not throw exception
         throw new RuntimeException("Unable to compute TOKENIZE schema.");
      }
   }
*/
/*
@Override
public Schema outputSchema(Schema input) {
   try{
      Schema.FieldSchema counter = new Schema.FieldSchema("dendrogram", DataType.CHARARRAY);
      Schema tupleSchema = new Schema(input.getField(0).schema.getField(0).schema.getFields());
      tupleSchema.add(counter);

      Schema.FieldSchema tupleFs;
      tupleFs = new Schema.FieldSchema("tuple", tupleSchema, DataType.TUPLE);

      Schema bagSchema = new Schema(tupleFs);
      return new Schema(new Schema.FieldSchema("bag",
                                                bagSchema, DataType.BAG));
     }catch (Exception e){
        return null;
     }
  } 
*/
}