/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
  /* Author: Zeehasham Rasheed
	Computer Science Department
	George Mason University 
	Dated: 10/15/2012
*/

package gmu.cs.dm.pig.eval;

import java.io.IOException;
import java.util.*;

import gmu.cs.dm.sequence.SequenceString;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.pig.EvalFunc;
import org.apache.pig.FuncSpec;
import org.apache.pig.data.*;
import org.apache.pig.impl.logicalLayer.schema.Schema;
import org.apache.pig.impl.logicalLayer.FrontendException;

/**
 * This function divides a search query string into wrods and extracts
 * n-grams with up to _ngramSizeLimit length.
 * Example 1: if query = "a real nice query" and _ngramSizeLimit = 2,
 * the query is split into: a, real, nice, query, a real, real nice, nice query
 * Example 2: if record = (u1, h1, pig hadoop) and _ngramSizeLimit = 2,
 * the record is split into: (u1, h1, pig), (u1, h1, hadoop), (u1, h1, pig hadoop)
 */
public class StringGenerator extends EvalFunc<DataBag> {
   private static final Log LOG = LogFactory.getLog(StringGenerator.class );

   public DataBag exec(Tuple input) throws IOException
   {
      DataBag output = DefaultBagFactory.getInstance().newDefaultBag();
       int start, num, end;
       boolean bothEnds = false;

      if ((input == null) || (input.size() == 0))
      {
         return(null);
      }

      try{
          String ba;
		  String id;
		  byte[] batemp;
		  byte[] idtemp;
          Object values = input.get(0);
		  batemp = ((DataByteArray) values).get();
          // i know this is somewhat inefficient, but quick and dirty
          ba = SequenceString.byteArrayToSequence(batemp);

          Tuple t = DefaultTupleFactory.getInstance().newTuple(2);
          t.set(0, new String(ba));
		  
		  id = (String)input.get(1);
		  t.set(1, new String(id));
          output.add(t);
      }
      catch (Exception e) {
         System.err.println("StringGenerator: failed to process input; error - " + e.getMessage());
         return(null);
      }
      return(output);
   }
   @Override
   public Schema outputSchema(Schema input)
   {
      try {
         Schema.FieldSchema tokenFs = new Schema.FieldSchema("seq",
                                                             DataType.CHARARRAY);
         //Schema tupleSchema = new Schema(tokenFs);
		
		Schema tupleSchema = new Schema();
        //tupleSchema.add(input.getField(0));
        //tupleSchema.add(input.getField(1));
        tupleSchema.add(tokenFs);
        tupleSchema.add(input.getField(1));
		
         Schema.FieldSchema tupleFs;
         tupleFs = new Schema.FieldSchema("tuple_of_seq", tupleSchema,
                                          DataType.TUPLE);

         Schema bagSchema = new Schema(tupleFs);
         bagSchema.setTwoLevelAccessRequired(true);
         Schema.FieldSchema bagFs = new Schema.FieldSchema(
            "seqs", bagSchema, DataType.BAG);

         return(new Schema(bagFs));
      }
      catch (FrontendException e) {
         // throwing RTE because
         //above schema creation is not expected to throw an exception
         // and also because superclass does not throw exception
         throw new RuntimeException("Unable to compute TOKENIZE schema.");
      }
   }


}
