/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
 
  /* Author: Zeehasham Rasheed
	Computer Science Department
	George Mason University 
	Dated: 10/15/2012
*/

package gmu.cs.dm.pig.eval;

import java.io.IOException;
import java.util.*;

import gmu.cs.dm.sequence.SequenceString;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.pig.EvalFunc;
import org.apache.pig.FuncSpec;
import org.apache.pig.data.*;
import org.apache.pig.impl.logicalLayer.schema.Schema;
import org.apache.pig.impl.logicalLayer.FrontendException;

/**
 * This function divides a search query string into wrods and extracts
 * n-grams with up to _ngramSizeLimit length.
 * Example 1: if query = "a real nice query" and _ngramSizeLimit = 2,
 * the query is split into: a, real, nice, query, a real, real nice, nice query
 * Example 2: if record = (u1, h1, pig hadoop) and _ngramSizeLimit = 2,
 * the record is split into: (u1, h1, pig), (u1, h1, hadoop), (u1, h1, pig hadoop)
 */
public class TranslateToKmer extends EvalFunc<DataBag> {
   private static final Log LOG = LogFactory.getLog(TranslateToKmer.class );

   public DataBag exec(Tuple input) throws IOException
   {
      //Tuple output = TupleFactory.getInstance().newTuple(1);
		DataBag output = DefaultBagFactory.getInstance().newDefaultBag();
      if ((input == null) || (input.size() == 0))
      {
         return(null);
      }

      try{
	  
		String SeqID = (String)input.get(1);
        String Seq   = (String)input.get(0);
		int KmerSize = (Integer)input.get(2);
		//System.out.println(SeqID);
		//System.out.println(Seq);
		//System.out.println(KmerSize);
		
        int SeqLength = Seq.length();
		
		long[] SeqKmerRep = new long[SeqLength-KmerSize+1];

		long kmer_prod;
		long kmer_sum;
		int k;
		char mychar;
		long charint;
		String SubSeq;
		
		int[] seqMap;
		seqMap = new int[256];
        seqMap['A'] = 0;
        seqMap['C'] = 1;
        seqMap['G'] = 2;
        seqMap['T'] = 3;
		seqMap['a'] = 0;
        seqMap['c'] = 1;
        seqMap['g'] = 2;
        seqMap['t'] = 3;
		
		
		Tuple t = DefaultTupleFactory.getInstance().newTuple(SeqLength-KmerSize+1);
		DataBag oo = DefaultBagFactory.getInstance().newDefaultBag();
         for (int i=0; i <SeqLength-KmerSize+1; i++)
         {
			if (i==SeqLength-KmerSize){
				SubSeq = Seq.substring(i);
				}
			else{	
				SubSeq = Seq.substring(i,i+KmerSize);
			}
			
			//System.out.println(SubSeq);
			
			for (kmer_prod=1,kmer_sum=0,k=0; k<KmerSize; k++){
			  //mychar = SubSeq.charAt(k);
			  //charint = mychar;
			  charint = seqMap[SubSeq.charAt(k)];

              kmer_sum += kmer_prod * (charint);
              kmer_prod = kmer_prod * 4;
            }
			SeqKmerRep[i] = kmer_sum;
			
			//Tuple t = DefaultTupleFactory.getInstance().newTuple(1);
			//t.set(i, kmer_sum);
		}

		//Arrays.sort(SeqKmerRep);
		
		for (int i=0; i <SeqKmerRep.length; i++){
			t.set(i, SeqKmerRep[i]);
		}
		oo.add(t);
		Tuple t2 = DefaultTupleFactory.getInstance().newTuple(2);
		t2.set(0,oo);
		
		//for(int i=0; i <SeqLength-KmerSize; i++){
		//System.out.println(SeqKmerRep[i]);
		//}
		
        //output.set(0, SeqKmerRep);
		
		//t.set(SeqLength-KmerSize, SeqID);
		DataBag ooo = DefaultBagFactory.getInstance().newDefaultBag();
		Tuple t3 = DefaultTupleFactory.getInstance().newTuple(1);
		t3.set(0, SeqID);
		ooo.add(t3);
		
		t2.set(1,ooo);
		output.add(t2);
		
		return(output);
      }
      catch (Exception e) {
         System.err.println("TranslateToKmer: failed to process input; error - " + e.getMessage());
         return(null);
      }
      
   }
   @Override
   public Schema outputSchema(Schema input)
   {
      try {
         Schema.FieldSchema tokenFs = new Schema.FieldSchema("seqkmer",
                                                             DataType.LONG);
         //Schema tupleSchema = new Schema(tokenFs);
		
		Schema tupleSchema = new Schema();
        //tupleSchema.add(input.getField(0));
        //tupleSchema.add(input.getField(1));
        tupleSchema.add(tokenFs);
        tupleSchema.add(input.getField(1));
		
         Schema.FieldSchema tupleFs;
         tupleFs = new Schema.FieldSchema("tuple_of_seq_kmer", tupleSchema,
                                          DataType.TUPLE);

         Schema bagSchema = new Schema(tupleFs);
         bagSchema.setTwoLevelAccessRequired(true);
         Schema.FieldSchema bagFs = new Schema.FieldSchema(
            "seqskmers", bagSchema, DataType.BAG);

         return(new Schema(bagFs));
      }
      catch (FrontendException e) {
         // throwing RTE because
         //above schema creation is not expected to throw an exception
         // and also because superclass does not throw exception
         throw new RuntimeException("Unable to compute TOKENIZE schema.");
      }
   }


}
